/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */

#ifndef QUB_BASICCURVE_H
#define QUB_BASICCURVE_H

// framework for most common curves -- see Linear for an example.

#include "qubfits.h"

typedef struct {
	double m, b;
} SampleComponent; // with N=2

#define MAX_NAMELEN 32

static char SampleNames[][MAX_NAMELEN] = {"m", "b"};

template<class ComponentType, int N>
class QUB_BasicCurve : public IQUB_Curve
{
protected:
	typedef ComponentType Component;

	const char *name;
	string eqn;
	char param_names[N][MAX_NAMELEN];
	Component initComp;
	IQUB_Curve *baseCurve;

	int ncomp;
	vector<Component> components;
	double *params; // alias
	vector<int> paramConst;

	QUB_Evts events;
	QUB_Tree state;
	vector<QUB_Tree> stateComponents;
	string tmpName;

	// to send a warning/error message, set "message" then send "Curve Msg" or "Curve Err" event
	string message;
public:
	// required:
	virtual void evalOne(QF_Data *data, double *yy, int component) = 0;
	virtual void evalAll(QF_Data *data, double *yy)
	{
		for (int c=0; c<ncomp; ++c)
			evalOne(data, yy, c);
	}

	// optional, if you have analytical partial derivatives:
	virtual int __cdecl hasDeriv() { return 0; }
	virtual void evalDeriv(QF_Data *data, double *dpar) {}
	
	// implement initParams() if you can guess reasonable starting values given data
	void __cdecl initParams(double *xx, double *yy, int n) {}

	// we give you x0:
	double getX0() { return getParamVal(getParamCount()-1); }

	// framework:

	QUB_BasicCurve(const char * curveName, const char * curveEqn, char *paramNames, Component initComponent, IQUB_Curve* baselineCurve)
		: name(curveName),
		  eqn(curveEqn),
		  initComp(initComponent),
		  baseCurve(baselineCurve),
		  ncomp(1),
		  components(ncomp),
		  params((double*) &(components[0])),
		  paramConst(ncomp*N, 0),
		  state(QUB_Tree::Create("Curve")),
		  stateComponents(ncomp)
	{
		memcpy(&(param_names[0][0]), paramNames, MAX_NAMELEN * N);
		components[0] = initComp;
		// populate the state tree
		// the only required element is "Name", and it must be the same as getName()
		state["Name"].setData( this->getName() );
		state["ComponentCount"].setData(QTR_TYPE_INT, ncomp);
		QUB_Tree comps = state["Components"];
		for (int c=0; c<ncomp; ++c) {
			QUB_Tree comp = comps.appendChild("Component");
			for (int i=0; i<N; ++i) {
				QUB_Tree par = comp[param_names[i]];
				par.setNumData(QTR_TYPE_DOUBLE, 2, 1, (void*)0);
				par.dataAs(0,0,(double)0) = params[c*N+i];
				par.dataAs(1,0,(double)0) = (double) paramConst[c*N+i];
			}
			stateComponents[c] = comp;
		}
		QUB_Tree baseState( baseCurve->getState() );
		QTR_DECREF( baseState.getImpl() );
		baseState.setName("Baseline");
		state.appendChild( baseState );

		if ( baseCurve->getEqn() != "" )
			eqn = eqn + string(" + ") + baseCurve->getEqn();
	}

	virtual ~QUB_BasicCurve()
	{
		delete baseCurve;
	}

	// for Delphi compatibility
	void __cdecl Free() { delete this; }

	IQUB_Evts * __cdecl Events() { return &events; }

	void __cdecl setExtraVars(char **names, int count) {}

	// When something is changed, we send the appropriate event.
	// qubfits provides a QUB_Evts object to dispatch events (below)

	// All set methods are requests, not commands.
	// For instance, we don't let users change the equation, so
	// setEqn() is blank.

	virtual const char * __cdecl getName() { return name; }
	virtual const char * __cdecl getEqn()  { return eqn.c_str(); }
	void __cdecl setEqn(const char *x) {}    // you probably don't want to allow this;
	int __cdecl canSetEqn() { return 0; }    // return 0 unless setEqn() works

	// to allow fitting to a sum of multiple curves:
	int __cdecl getComponentCount()        { return ncomp; }
	void __cdecl setComponentCount(int x)
	{
		if ( (x > 0) && (x != ncomp) ) {
			QUB_Tree comps = state["Components"];
			// remove excess components
			for (int i=x; i<ncomp; ++i) {
				comps.removeChild(stateComponents[i]);
			}
			components.resize(x);
			stateComponents.resize(x);
			params = (double*) & (components[0]);
			paramConst.resize(N*x);
			// add components
			for (int c=ncomp; c<x; ++c) {
				components[c] = initComp;
				QUB_Tree comp = comps.appendChild("Component");
				for (int i=0; i<N; ++i) {
					QUB_Tree par = comp[param_names[i]];
					par.setNumData(QTR_TYPE_DOUBLE, 2, 1, (void*)0);
					par.dataAs(0,0,(double)0) = params[c*N+i];
					par.dataAs(1,0,(double)0) = (double) paramConst[c*N+i];
				}
				stateComponents[c] = comp;
			}
			ncomp = x;
			state["ComponentCount"].setData(QTR_TYPE_INT, ncomp);
			events.send("Curve CompCount", this);
			events.send("Curve Params", this);
		}
	}

	// curve parameters:
	// if you have 2 basic parameters, e.g. m and b,
	// and 3 components,
	// you should list 6 params: m1,b1,m2,b2,m3,b3
	int __cdecl getParamCount()
	{
		return ( (ncomp * N) + baseCurve->getParamCount() );
	}
	const char * __cdecl getParamName(int ix)
	// we build a string on the fly, and store it in an instance variable
	// so it still exists after return
	{
		if ( (ix >= 0) && (ix < getParamCount()) ) {
			if ( (ncomp == 1) && (ix < (ncomp * N)) )
				return param_names[ix];

			int iname = ix % N;
			int icomp = ix / N;
			if ( icomp < ncomp ) {
				ostringstream ost;
				ost << param_names[iname] << (icomp+1);
				tmpName = ost.str();
				return tmpName.c_str();
			}
			else {
				return baseCurve->getParamName(ix - (ncomp*N));
			}
		}
		return "";
	}
	double __cdecl getParamVal(int ix)
	{
		if ( ix < (ncomp * N) )
			return params[ix];
		else
			return baseCurve->getParamVal(ix - (ncomp*N));
	}
	void __cdecl setParamVal(int ix, double val, int sendEvent)
	// sendEvent could be 0 (false) if all params are being set, one at a time
	{
		if ( (ix >= 0) && (ix < getParamCount())) {
			if ( ix < (ncomp * N) ) {
				params[ix] = val;
				stateComponents[ix / N][param_names[ix % N]].dataAs(0,0,(double)0) = val;
			}
			else {
				baseCurve->setParamVal(ix - (ncomp*N), val, 0);
			}
		}
		if ( sendEvent )
			events.send("Curve ParamVal", this);
	}
	int __cdecl getParamConst(int ix)
	{
		if ( ix < (ncomp * N) )
			return paramConst[ix];
		else
			return baseCurve->getParamConst(ix - (ncomp*N));
	}
	void __cdecl setParamConst(int ix, int isConst)
	{
		if ( ix < (ncomp * N) ) {
			paramConst[ix] = isConst;
			stateComponents[ix / N][param_names[ix % N]].dataAs(1,0,(double)0) = (double) isConst;
		} else {
			baseCurve->setParamConst(ix - (ncomp*N), isConst);
		}
		events.send("Curve ParamConst", this);
	}
	void __cdecl evalArrayAll(QF_Data *data, double *yy)
	{
                baseCurve->evalArrayAll(data, yy);
                evalAll(data, yy);
                clearfp();
	}
	void __cdecl evalArrayOne(QF_Data *data, double *yy, int component)
	{
		baseCurve->evalArrayAll(data, yy);
		evalOne(data, yy, component);
                clearfp();
	}
	void __cdecl evalDerivArray(QF_Data *data, double *dpar)
	{
		int localNpar = ncomp * N;
		baseCurve->evalDerivArray(data, dpar+localNpar);
		for (int i=0; i<localNpar; ++i)
			dpar[i] = 0.0;
		evalDeriv(data, dpar);
                clearfp();
	}


	const char * __cdecl getMessage()
	{
		return message.c_str();
	}

	// for the presets system, you need to keep your state updated in a qubtree
	// (qubtree: http://www.qub.buffalo.edu/qubdoc)
	QTR_Impl* __cdecl getState()
	{
		QTR_Impl* pstate = state.getImpl();
		QTR_INCREF(pstate);
		return pstate;
	}
	// and update your state from a "preset" (stored copy of the state tree)
	void __cdecl updateState(QTR_Impl* pustate)
	{
		QUB_Tree ustate(pustate);
		baseCurve->updateState( ustate.find("Baseline")->getImpl() );
		QUB_Tree sub;
		if ( ustate["Name"].dataAsString() == this->getName() ) {
			sub = * ustate.find("ComponentCount");
			if ( sub.dataCount() > 0 )
				this->setComponentCount( sub.dataAsInt(1) );
			int ipar = 0;
			for ( QUB_TreeMonoIter ci = ustate["Components"].find("Component");
				! ci->isNull(); ci.nextSameName() ) {
			  for (int i=0; i<N; ++i) {
				sub = * ci->find(param_names[i]);
				if ( sub.dataCount() > 0 )
					this->setParamVal(ipar++, sub.dataAs(0,0,(double)0), 0);
				if ( sub.dataCount() > 1 )
					this->setParamConst(ipar-1, (int)sub.dataAs(1,0,(double)0));
			  }
			}
			this->setParamVal(0, this->getParamVal(0), 1); // trigger event
		}
	}
	void __cdecl clearfp()
	{
#if defined(_WIN32) 
		unsigned int fstat = _clearfp(); // just in case
		_fpreset();
#endif
	}
};

// make your own factory function with this signature,
// returning a new instance of your class:

// vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv
// IQUB_Curve* __stdcall CreateLinear();
// ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
	

// baseline choices: none: CreateNullCurve()
//                   constant: CreateConst()
//                   linear: CreateLinear()


QUBFITS_API IQUB_Curve* __stdcall CreateConst();
QUBFITS_API IQUB_Curve* __stdcall CreateLinear();


#endif
